--Xfade pow v0.1 by Aftab Hussain aka afta8 - fathand@gmail.com - 15th July 2014

--[[============================================================================
main.lua
============================================================================]]--

-- Placeholder for the dialog
local dialog = nil

-- Placeholder to expose the ViewBuilder outside the show_dialog() function
local vb = nil

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
  
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value


--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------

-- Function to do a crossfade on a selected sample 
local function do_crossfade(smp, loop_start, loop_end)

  -- Setup
  local smp_buffer = smp.sample_buffer
  local num_channels = smp_buffer.number_of_channels 
  local sample_size = smp_buffer.number_of_frames
   
  -- Loop size as a percentage of sample size
  local crossfade_loop_size = ((loop_end - loop_start)/sample_size)*100

  -- Variables to track fade in and fade out start positions (frames)
  local fade_in_start_pos = nil
  local fade_out_start_pos = nil

  -- Size of area being processed (frames)
  local process_area_size = nil

  -- Math
  local sin = math.sin
  local cos = math.cos
  local constant = 2*(math.pi/4)
    
  -- Functions for doing the fades  
  -- Fade in function - Returns faded value for a given frame
  local function fade_in(frm, val, multiplier)    
    -- Equal power fade
    local pos = (frm*multiplier)/100
    local angle = pos*constant   
    return sin(angle)*val
  end
    
  -- Fade out function - Returns faded value for a given frame  
  local function fade_out(frm, val, multiplier)
    -- Equal power fade
    local pos = (frm*multiplier)/100
    local angle = pos*constant    
    return cos(angle)*val
  end  


  --[[ Sample processing ]]--
  local function process(channel)

    -- Sample data buffers
    local sample_data = { }
    local fade_in_sample_data = { }

    -- Set up fade in and out start positions and also size of processed area  
    if crossfade_loop_size <= 50 then
      -- Set fade in out positions if loop point is 50% or less of sample size
      process_area_size = loop_end - loop_start
      fade_in_start_pos = loop_start - process_area_size + 1
      fade_out_start_pos = loop_start                
    else
      -- Fade in out positions if loop point is more than 50% of sample size
      process_area_size = loop_start
      fade_in_start_pos = 1
      fade_out_start_pos = loop_end - process_area_size                  
    end
    
    -- Read channel into buffers
    for n = 1, loop_end, 1 do
      sample_data[n] = smp_buffer:sample_data(channel, n)
    end

    for n = 1, loop_end, 1 do
      fade_in_sample_data[n] = smp_buffer:sample_data(channel, n)
    end

    -- % volume scale 
    local multiplier = 100/process_area_size -- Sets the multiplier for the fade rate
          
    -- Set up process counter
    local proc_count = nil

    -- Do fade in processing
    proc_count = 1
    for n = fade_in_start_pos, loop_start, 1 do
      fade_in_sample_data[n] = fade_in(proc_count, fade_in_sample_data[n], multiplier)
      proc_count = proc_count + 1
    end
    
    -- Do fade out processing 
    proc_count = 1
    for n = fade_out_start_pos, loop_end, 1 do
      sample_data[n] = fade_out(proc_count, sample_data[n], multiplier)
      proc_count = proc_count + 1
    end
    
    -- Create crossfaded sample data
    local a = nil
    local b = nil
    proc_count = fade_in_start_pos
    for n = fade_out_start_pos, loop_end, 1 do
      a = sample_data[n]
      b = fade_in_sample_data[proc_count]
      sample_data[n] = a + b
      proc_count = proc_count + 1
    end
  
    -- Writing to sample buffer
    for n = fade_out_start_pos, loop_end, 1 do
      smp_buffer:set_sample_data(channel, n, sample_data[n])
    end
    
  end

  -- Process audio channels
  if num_channels > 1 then
    process(1)
    process(2)
  else
    process(1)  
  end

end


-- Main function
local function xfadepow(loop_size)

  local smp = renoise.song().selected_sample
  local smp_buffer = smp.sample_buffer
  local sample_size = smp_buffer.number_of_frames 
    
  -- Get loop start position (frames)
  local loop_start = math.floor((sample_size*loop_size)+0.5)
  
  -- Setup loop
  smp.loop_mode = 2
  smp.loop_start = loop_start
  smp.loop_end = sample_size  
    
  -- Do crossfade
  smp_buffer:prepare_sample_data_changes()
  do_crossfade(smp, loop_start, sample_size)    
  smp_buffer:finalize_sample_data_changes()

end


--------------------------------------------------------------------------------
-- GUI
--------------------------------------------------------------------------------

local function show_dialog()

  -- This block makes sure a non-modal dialog is shown once.
  -- If the dialog is already opened, it will be focused.
  if dialog and dialog.visible then
    dialog:show()
    return
  end
  
  -- The ViewBuilder is the basis
  vb = renoise.ViewBuilder()
  
  -- Declare variables
  local CONTENT_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  local DEFAULT_DIALOG_BUTTON_HEIGHT = renoise.ViewBuilder.DEFAULT_DIALOG_BUTTON_HEIGHT
  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN

  -- Default settings
  local loop_size = 0.5
  
  -- Crossfade loop size
  local loop_size_row = vb:row {
    
    vb:minislider {
      id = "crossfadeloopsize",
      active = true,
      width = 400,
      height = 20,
      min = 0,
      max = 1,
      value = 0.5,
      tooltip = "The start position of the loop",
      notifier = function(value)
        loop_size = value
      end
    }
  }

  -- Process button
  local process_button_row = vb:horizontal_aligner {
    mode = "center",
    
    vb:button {
      text = "Create Xfade Loop",
      height = 25,
      width = 55,
      
      notifier = function()
          xfadepow(loop_size)
      end
    }        
  } 


  -- Setting up the GUI layout
  local content = vb:column {
    margin = DIALOG_MARGIN/1.5,
    spacing = CONTENT_SPACING,
    
    vb:column{
      spacing = 2*CONTENT_SPACING,
              
      vb:column {
        spacing = CONTENT_SPACING,
        style = "group",
        margin = DIALOG_MARGIN/1.5,
        
 
        vb:column {
          id = "looppanel",
          margin = DIALOG_MARGIN/1.5,
          spacing = CONTENT_SPACING,
          style = "border",

          loop_size_row,
        } 
      },
      process_button_row
    }  
  }    


  -- Key passthrough function
  local function keyhandler(dialog, key)
    return key
  end

      
  -- A custom dialog is non-modal and displays a user designed
  -- layout built with the ViewBuilder.   
  dialog = renoise.app():show_custom_dialog("Equal Power Xfade Loop", content, keyhandler)  
  
end


--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:"..tool_name.."...",
  invoke = show_dialog  
}

--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:" .. tool_name.."...",
  invoke = show_dialog
}

--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_dialog
}
--]]
